/**********************************************************************
 *** implantation de RC4 et fonction pour modéliser le codage WEP     *
 *** pour le TP3 d'info-614 : RC4, WEP et attaque FMS                 *
 *** Pierre Hyvernat                                                  *
 **********************************************************************/

#include "tp3.h"

/**********************************************************************
 * Le code pour RC4 :                                                 *
 *   - initialisation (RC4_KSA())                                     *
 *   - obtention d'un nouvel octet (RC4_PRGA())                       *
 * et une fonction pour générer une clé secrète aléatoire.            *
 **********************************************************************/

/* état interne du générateur */
typedef struct etat_RC4 {
  /* la permutation */
  octet S[256];

  /* les compteurs i et j */
  octet i;
  octet j;
} etat_RC4;

/* Une variable (globale et statique) contenant l'état courant du générateur */
static etat_RC4 etat;

/* la taille de la clé */
/* ATTENTION : pour les clés WEP, la taille comprends les octets des vecteurs
 * d'initialisation. Pour le WEP normal, "taille_cle" est donc de 16, même si
 * la clé WEP elle même ne fait que 13 octets. */
int taille_cle = 16;

/* Une variable (globale et statique) contenant la clé secrète */
/* ATTENTION : dans l'application pour les clés WEP, les 3 premiers octets ne
 * servent à rien : ils sont remplacés par le vecteur d'initialisation */
static octet *cle_secrete = NULL;

/* fonction auxiliaire : échange deux octets */
void echange(octet *a, octet *b)
{
  octet o;
  o = *a;
  *a = *b;
  *b = o;
}

/**********************************************************************
 * RC4, le "key scheduling algorithm" : initialisation de l'état      *
 **********************************************************************/
void RC4_KSA(void)
{
  int i, j;

  /* on remet les compteurs à 0 */
  etat.i = 0;
  etat.j = 0;

  /* la permutation S est initialement l'identité */
  for (i = 0; i < 256; i++)
    etat.S[i] = i;

  /* le mélange pour obtenir la permutation... */
  j = 0;
  for (i = 0; i < 256; i++) {
    j = (j + etat.S[i] + cle_secrete[i % taille_cle]) % 256;
    echange(&etat.S[i], &etat.S[j]);
  }
}

/**********************************************************************
 * RC4, générer un nouvel octet à partir de l'état courant.           *
 **********************************************************************/
octet RC4_PRGA(void)
{
  octet k;

  etat.i += 1;
  etat.j += etat.S[etat.i];
  echange(&etat.S[etat.i], &etat.S[etat.j]);

  k = (etat.S[etat.i] + etat.S[etat.j]);
  return etat.S[k];
}

/**********************************************************************
 * génère une clé secrète aléatoire                                   *
 **********************************************************************/
void nouvelle_cle_aleatoire()
{
  cle_secrete = realloc(cle_secrete, taille_cle * sizeof(octet));
  for (int i = 0; i < taille_cle; i++)
    cle_secrete[i] = rand() % 256;
}

/**********************************************************************
 * Quelques fonctions pour tester RC4                                 *
 **********************************************************************/

/* chiffre une suite d'octets en initialisant RC4 à partir d'une clé :
 *   - "l" est la taille de la clé,
 *   - "cle" est la clé secrète
 *   - "n" est la taille du message clair
 *   - "texte" est le message clair
 */
octet *test_chiffre(int l, octet *cle, int n, octet *texte)
{

  int i;

  taille_cle = l;
  cle_secrete = realloc(cle_secrete, sizeof(octet) * l);
  for (i = 0; i < l; i++)
    cle_secrete[i] = cle[i];

  octet *resultat = malloc(n * sizeof(octet));

  RC4_KSA();
  for (i = 0; i < n; i++)
    resultat[i] = texte[i] ^ RC4_PRGA();

  return resultat;
}

/* chiffre une chaîne de caractères en initialisant RC4 à partir d'une clé,
 * elle même donnée comme chaîne de caractères :
 *   - "cle" est la clé secrète
 *   - "texte" est le message clair
 */
octet *chiffre_chaine(char *cle, char *texte)
{
  return (test_chiffre (strlen(cle), (octet *) cle, strlen(texte), (octet *) texte));
}

/* idem, mais vérifie si le résultat est bien celui attendu
 * renvoie 1  si le test est OK, et "-n" si le test a échoué sur l'octet "n" */
int test_chaine(char *cle, char *texte, octet *resultat)
{
  octet *rc4 = chiffre_chaine(cle, texte);
  int OK = 1;
  for (int i = 0; i < (int)strlen(texte); i++) {
    if (rc4[i] != resultat[i]) {
      OK = -i;
      break;
    }
  }
  free(rc4);
  return OK;
}

void test_RC4(void)
{

  printf("\n");
  printf("################\n");
  printf("# TESTS RC4... #\n");
  printf("################\n\n");

  {
    char *cle = "Key";
    char *texte = "Plaintext";
    octet R[] = { 0xbb, 0xf3, 0x16, 0xe8, 0xd9, 0x40, 0xaf, 0x0a, 0xd3 };
    int OK;
    printf("Test 1 : cle = \"%s\" et texte = \"%s\" :\n", cle, texte);
    OK = test_chaine(cle, texte, R);
    if (OK == 1) {
      printf("  OK.\n");
    } else {
      printf("  OOPS : erreur à l'octet %d.\n", -OK);
    }
  }
  {
    char *cle = "Wiki";
    char *texte = "pedia";
    octet R[] = { 0x10, 0x21, 0xbf, 0x04, 0x20 };
    int OK;
    printf("Test 2 : cle = \"%s\" et texte = \"%s\" :\n", cle, texte);
    OK = test_chaine(cle, texte, R);
    if (OK == 1) {
      printf("  OK.\n");
    } else {
      printf("  OOPS : erreur à l'octet %d.\n", -OK);
    }
  }
  {
    char *cle = "Secret";
    char *texte = "Attack at dawn";
    octet R[] = { 0x45, 0xA0, 0x1F, 0x64, 0x5F, 0xC3, 0x5B, 0x38, 0x35, 0x52, 0x54, 0x4B, 0x9B, 0xF5 };
    int OK;
    printf("Test 3 : cle = \"%s\" et texte = \"%s\" :\n", cle, texte);
    OK = test_chaine(cle, texte, R);
    if (OK == 1) {
      printf("  OK.\n");
    } else {
      printf("  OOPS : erreur à l'octet %d.\n", -OK);
    }
  }

}

/**********************************************************************
 * fonction pour modéliser l'utilisation de RC4 pour le WEP :         *
 * on concatène un vecteur d'initialisation ("IV") devant une clé     *
 * secrète...                                                         *
 **********************************************************************/

/* taille des vecteurs d'initialisation pour le WEP */
int taille_IV = 3;

/* Une variable globale contenant la clé secrète WEP */
octet *cle_wep = NULL;

/* initialise une clé secrète à partir d'un vecteur d'initialisation et de la
 * clé WEP */
void init_WEP_RC4(octet *IV)
{
  int c;
  for (c = 0; c < taille_IV; c++)
    cle_secrete[c] = IV[c];
  for (c = taille_IV; c < taille_cle; c++)
    cle_secrete[c] = cle_wep[c - taille_IV];
  RC4_KSA();
}

/* génère une clé WEP aléatoire */
void nouvelle_cle_wep_aleatoire()
{
  if (taille_cle <= taille_IV) {
    printf("Problème : les vecteurs d'initialisations sont plus grands que les clés !\n");
    printf("On arrête tout...\n");
    exit(-1);
  }
  cle_secrete = realloc(cle_secrete, taille_cle * sizeof(octet));
  cle_wep = realloc(cle_wep, (taille_cle - taille_IV) * sizeof(octet));
  for (int i = 0; i < taille_cle - taille_IV; i++)
    cle_wep[i] = rand() % 256;
}
